#include <iostream>
#include <list>
#include <cstdlib>

#include "edgelist.hpp"

using namespace std;

// aloca um novo grafo
Edgelist::Edgelist(int num_nodes)
{
  number_of_nodes = num_nodes;
  number_of_edges = 0;
  source = sink = -1;
}

Edgelist::Edgelist()
{
  number_of_nodes = 0;
  number_of_edges = 0;
  source = sink = -1;
}

Edge Edgelist::add_edge(int tail, int head, long cap)
{  
  Edge e = {tail, head, cap};

  set<Edge>::iterator it = edges.find(e);
  if (it == edges.end())  // insert new edge
    {
      edges.insert(e);
      number_of_edges++;
      return e;
    }
  (*it).cap += cap;  // increase the existing edge capacity
  return *it;
}
  
int Edgelist::read_dimacs(ifstream &fp)
{
  string line;
  int n1, n2, m;
  long cap;
  char c;

  while (fp.good()) {
    getline(fp, line);
    if(line[0] == 'p')
      break;
  };
  if (!fp.good())
    return -1;

  sscanf(line.c_str(), "%*c %*s %d %d", &number_of_nodes, &m);   
  while(getline(fp, line).good()){
    int count = sscanf(line.c_str(), "%c %d %d %ld", &c, &n1, &n2, &cap);     
    if (count == 3)
      if (n1 == 0)
	source = n2-1;
      else if (n1 == 1)
	sink = n2-1;
    if (count != 4 && count != 3) 
      {
	printf("Edglist::read_dimacs(): wrong number of entries in a line.\n");
	exit(1);
      }
    if (count == 4 && c != 'c') 
      add_edge(n1-1, n2-1, cap);
  }
  /*   if (number_of_edges != m)
    {
      cout << number_of_nodes << " vs. " << m << endl;
      cout << "Edgelist::read_dimacs(): wrong number of edges in the input." << endl;
      exit(1);
      } */
  return 0;
}

ostream& operator<<(ostream& os, Edgelist& G)
{
  os <<  "c graph" << endl;
  os << "p max " << G.num_nodes() << " " << G.num_edges() << endl;
 
  os << "n " << G.get_source() << " s" << endl; 
  os << "n " << G.get_sink()   << " t" << endl;

  Edgelist::edge_iterator it;
  for (it = G.edges.begin(); it != G.edges.end(); ++it)
    {
      Edge edge = *it;
      /* if (edge->tail < edge->head) */
      os << "a " << edge.tail << " " << edge.head << " " << edge.cap << endl;
    }
  return os;
}

void Edgelist::clear()
{
  edges.clear();
  number_of_nodes = number_of_edges = 0;
  source = sink = -1;
}
