#pragma once

// version that maintains the graph simple using a set data structure
// multiedgelist.hpp allows multiple edges using list instead of set

#include <fstream>
#include <set>

using namespace std;

/* classical adjacency list implementation */
/*
typedef struct st_edge {
  int v;   // outro vertice
  float w;  // peso da aresta
    struct st_edge *next;
} * edge;

typedef struct st_graph {
    edge *adj;
    int *grau;
    int n;
    int m;
} * graph;
*/

/* edge list representation */
struct  Edge {
  int tail;
  int head;
  // cap is mutable in order to be changed inplace in a set<Edge>
  // the < operator cannot use cap.
  mutable long cap; 

  /*  bool operator==(const Edge& other) const {
    return (this->tail == other.tail && this->head == other.head) || 
      (this->tail == other.head && this->head == other.tail);
  } 
  bool operator!=(const Edge& other) const {
    return !(*this == other); 
    } */ 
  bool operator<(const Edge& rhs) const {
    return ((*this).tail < rhs.tail) || ((*this).tail == rhs.tail && (*this).head < rhs.head); }
  /*  bool operator>(const Edge& rhs) const {
    if (*this == rhs)
      return false;
      return ((*this).tail > rhs.tail) || ((*this).tail == rhs.tail && (*this).head > rhs.tail); } */
};

class Edgelist {
  int number_of_nodes, 
      number_of_edges;    // number of vertices/edges 
  int source, sink; // source/sink if it is a max flow instance
  set<Edge> edges;  

 public:  
  Edgelist(int n);
  Edgelist();
  
  Edge add_edge(int tail, int head, long cap);
  int read_dimacs (ifstream& fp);
  //  void print();
  int num_nodes() { return number_of_nodes; }
  int num_edges() { return number_of_edges; }
  void set_size(int n) { number_of_nodes = n; };
  int get_source() { return source; }
  int get_sink() { return sink; }
  void set_source(int src) { source = src; }
  void set_sink(int snk) { sink = snk; }
  void clear();

  typedef std::set<Edge>::iterator edge_iterator;
  edge_iterator begin() { return edges.begin(); }
  edge_iterator end() { return edges.end(); }
  friend ostream& operator<<(ostream& os, Edgelist& el);
};




