/* Maximum flow - highest lavel push-relabel algorithm */
/* COPYRIGHT C 1995, 2000 by IG Systems, Inc., igsys@eclipse.net */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <iostream>
#include <cstdlib>
#include <utility>
#include <vector>
#include <math.h>
// #include <unordered_set>
#include <boost/unordered_set.hpp>
#include <boost/foreach.hpp>

#include "edgelist.hpp"

#include "values.h"
#include "types.h"          /* type definitions */
#include "timer.c"          /* timing routine */
#include "hi_pr.hpp"

/*
#define GLOB_UPDT_FREQ 0.5
*/
#define GLOB_UPDT_FREQ 0.5
#define ALPHA 6
#define BETA 12

#define WHITE 0
#define GREY 1
#define BLACK 2

/* global variables */

long   n;                    /* number of nodes */
long   m;                    /* number of arcs */
long   nm;                   /* n + ALPHA * m */
long   nMin;                 /* smallest node id */
node   *nodes;               /* array of nodes */
arc    *arcs;                /* array of arcs */
bucket *buckets;             /* array of buckets */
cType  *cap;                 /* array of capacities */
node   *source;              /* source node pointer */
node   *sink;                /* sink node pointer */
//node   **queue;              /* queue for BFS */
//node   **qHead, **qTail, **qLast;     /* queue pointers */
long   dMax;                 /* maximum label */
long   aMax;                 /* maximum actie node label */
long   aMin;                 /* minimum active node label */
double flow;                 /* flow value */
long pushCnt  = 0;           /* number of pushes */
long relabelCnt   = 0;       /* number of relabels */
long updateCnt    = 0;       /* number of updates */
long gapCnt   = 0;           /* number of gaps */
long gNodeCnt = 0;           /* number of nodes after gap */  
float t, t2;                 /* for saving times */
node   *sentinelNode;        /* end of the node list marker */
arc *stopA;                  /* used in forAllArcs */
long workSinceUpdate=0;      /* the number of arc scans since last update */
float globUpdtFreq;          /* global update frequency */

long   scc_n;                /* number of nodes in the scc graph */
long   scc_m;                /* number of arcs in the scc graph */
scc_node   *scc_nodes;           /* array of nodes of the scc graph */
scc_arc    *scc_arcs;            /* array of arcs of the scc graph */
scc_node   *scc_source;          /* scc graph node that contains the source */
scc_node   *scc_sink;            /* scc graph node that contains the sink   */
int scc_source_index;
int scc_sink_index;
scc_node   *scc_sentinelNode;    /* end of the node array                   */
typedef boost::unordered_set<scc_node*> pivot_set_t;
typedef boost::unordered_set<scc_nodeSt*>::iterator pivot_set_iterator;
float most_balanced_value;
int number_of_cuts; 
int S_size;         // keeps track of the number of (original) vertices in S

#include "parser.cpp"         /* parser */

/* macros */

#define forAllNodes(i) for ( i = nodes; i != sentinelNode; i++ )
#define forAllArcs(i,a) for (a = i->first, stopA = (i+1)->first; a != stopA; a++)

#define nNode( i ) ( (i) - nodes + nMin )
/* pointer to index */
#define Index( i ) ( (i) - nodes ) 

#define nArc( a )  ( ( a == NULL )? -1 : (a) - arcs )

#define min( a, b ) ( ( (a) < (b) ) ? a : b )

/* FIFO queue for BFS macros */
/*
#define qInit() \
{\
  qHead = qTail = queue;\
}

#define qEmpty ( qHead == qTail )

#define qEnqueue(i) \
{\
  *qTail = i;\
  if ( qTail == qLast ) qTail = queue;\
  else qTail++;\
}

#define qDequeue(i) \
{\
  i = *qHead;\
  if ( qHead == qLast ) qHead = queue;\
  else qHead++;\
}
*/

/* 
   bucket macros:
   bucket's active node list is singly-linked
     operations aAdd, aRemove (from the front)
   bucket's inactive list is doubly-linked
     operations iAdd, iDelete (from arbitrary position)
*/

long i_dist;

#define aAdd(l,i)\
{\
  i->bNext = l->firstActive;\
  l->firstActive = i;\
  i_dist = i->d;\
  if (i_dist < aMin)\
    aMin = i_dist;\
  if (i_dist > aMax)\
    aMax = i_dist;\
  if (dMax < aMax)\
    dMax = aMax;\
}

/* i must be the first element */
#define aRemove(l,i)\
{\
  l->firstActive = i->bNext;\
}

node *i_next, *i_prev;
#define iAdd(l,i)\
{\
  i_next = l->firstInactive;\
  i->bNext = i_next;\
  i->bPrev = sentinelNode;\
  i_next->bPrev = i;\
  l->firstInactive = i;\
}

#define iDelete(l,i)\
{\
  i_next = i->bNext;\
  if (l->firstInactive == i) {\
    l->firstInactive = i_next;\
    i_next->bPrev = sentinelNode;\
  }\
  else {\
    i_prev = i->bPrev;\
    i_prev->bNext = i_next;\
    i_next->bPrev = i_prev;\
  }\
}

/* allocate datastructures, initialize related variables */

int allocDS( )

{

  nm = ALPHA * n + m;
  /*
  queue = (node**) calloc ( n, sizeof (node*) );
  if ( queue == NULL ) return ( 1 );
  qLast = queue + n - 1;
  qInit();
  */
  // moved to hi_pr_init() - to alloc once
  // buckets = (bucket*) calloc ( n+2, sizeof (bucket) );
  // if ( buckets == NULL ) return ( 1 );

  sentinelNode = nodes + n;
  sentinelNode->first = arcs + 2*m;

  return ( 0 );

} /* end of allocate */


void init( )

{
  node  *i;        /* current node */
  int overflowDetected;
  bucket *l;
  arc *a;
#ifdef EXCESS_TYPE_LONG
  double testExcess;
#endif
#ifndef OLD_INIT
  unsigned long delta;
#endif

  // initialize excesses

  forAllNodes(i) {
    i->excess = 0;
    i->current = i->first;
    forAllArcs(i, a)
      a->resCap = cap[a-arcs];
  }

  for (l = buckets; l <= buckets + n-1; l++) {
    l -> firstActive   = sentinelNode;
    l -> firstInactive  = sentinelNode;
  }
    
  overflowDetected = 0;
#ifdef EXCESS_TYPE_LONG
  testExcess = 0;
  forAllArcs(source,a) {
    if (a->head != source) {
      testExcess += a->resCap;
    }
  }
  if (testExcess > MAXLONG) {
    printf("c WARNING: excess overflow. See README for details.\nc\n");
    overflowDetected = 1;
  }
#endif
#ifdef OLD_INIT
  source -> excess = MAXLONG;
#else
  if (overflowDetected) {
    source -> excess = MAXLONG;
  }
  else {
    source->excess = 0;
    forAllArcs(source,a) {
      if (a->head != source) {
	pushCnt ++;
	delta = a -> resCap;
	a -> resCap -= delta;
	(a -> rev) -> resCap += delta;
	a->head->excess += delta;
      }
    }
  }

  /*  setup labels and buckets */
  l = buckets + 1;
    
  aMax = 0;
  aMin = n;
    
  forAllNodes(i) {
    if (i == sink) {
      i->d = 0;
      iAdd(buckets,i);
      continue;
    }
    if ((i == source) && (!overflowDetected)) {
      i->d = n;
    }
    else
      i->d = 1;
    if (i->excess > 0) {
      /* put into active list */
      aAdd(l,i);
    }
    else { /* i -> excess == 0 */
      /* put into inactive list */
      if (i->d < n)
	iAdd(l,i);
    }
  }
  dMax = 1;
#endif

  //  dMax = n-1;
  //  flow = 0.0;

} /* end of init */

void checkMax()

{
  bucket *l;

  for (l = buckets + dMax + 1; l < buckets + n; l++) {
    assert(l->firstActive == sentinelNode);
    assert(l->firstInactive == sentinelNode);
  }
}

/* global update via backward breadth first search from the sink */

void globalUpdate ()

{

  node  *i, *j;       /* node pointers */
  arc   *a;           /* current arc pointers  */
  bucket *l, *jL;          /* bucket */
  long curDist, jD;
  long state;


  updateCnt ++;

  /* initialization */

  forAllNodes(i)
    i -> d = n;
  sink -> d = 0;

  for (l = buckets; l <= buckets + dMax; l++) {
    l -> firstActive   = sentinelNode;
    l -> firstInactive  = sentinelNode;
  }

  dMax = aMax = 0;
  aMin = n;

  /* breadth first search */

  // add sink to bucket zero

  iAdd(buckets, sink);
  for (curDist = 0; 1; curDist++) {

    state = 0;
    l = buckets + curDist;
    jD = curDist + 1;
    jL = l + 1;
    /*
    jL -> firstActive   = sentinelNode;
    jL -> firstInactive  = sentinelNode;
    */

    if ((l->firstActive == sentinelNode) && 
	(l->firstInactive == sentinelNode))
      break;

    while (1) {

      switch (state) {
      case 0: 
	i = l->firstInactive;
	state = 1;
	break;
      case 1:
	i = i->bNext;
	break;
      case 2:
	i = l->firstActive;
	state = 3;
	break;
      case 3:
	i = i->bNext;
	break;
      default: 
	assert(0);
	break;
      }
      
      if (i == sentinelNode) {
	if (state == 1) {
	  state = 2;
	  continue;
	}
	else {
	  assert(state == 3);
	  break;
	}
      }

      /* scanning arcs incident to node i */
      forAllArcs(i,a) {
	if (a->rev->resCap > 0 ) {
	  j = a->head;
	  if (j->d == n) {
	    j->d = jD;
	    j->current = j->first;
	    if (jD > dMax) dMax = jD;
	    
	    if (j->excess > 0) {
	      /* put into active list */
	      aAdd(jL,j);
	    }
	    else {
	      /* put into inactive list */
	      iAdd(jL,j);
	    }
	  }
	}
      } /* node i is scanned */ 
    }
  }

} /* end of global update */

/* second stage -- preflow to flow */
void stageTwo ( )
/*
   do dsf in the reverse flow graph from nodes with excess
   cancel cycles if found
   return excess flow in topological order
*/

/*
   i->d is used for dfs labels 
   i->bNext is used for topological order list
   buckets[i-nodes]->firstActive is used for DSF tree
*/

{
  node *i, *j, *tos, *bos, *restart, *r;
  arc *a;
  cType delta;

  /* deal with self-loops */
  forAllNodes(i) {
    forAllArcs(i,a)
      if ( a -> head == i ) {
	a -> resCap = cap[a - arcs];
      }
  }

  /* initialize */
  tos = bos = NULL;
  forAllNodes(i) {
    i -> d = WHITE;
    //    buckets[i-nodes].firstActive = NULL;
    buckets[i-nodes].firstActive = sentinelNode;
    i -> current = i -> first;
  }

  /* eliminate flow cycles, topologicaly order vertices */
  forAllNodes(i)
    if (( i -> d == WHITE ) && ( i -> excess > 0 ) &&
	( i != source ) && ( i != sink )) {
      r = i;
      r -> d = GREY;
      do {
	for ( ; i->current != (i+1)->first; i->current++) {
	  a = i -> current;
	  if (( cap[a - arcs] == 0 ) && ( a -> resCap > 0 )) { 
	    j = a -> head;
	    if ( j -> d == WHITE ) {
	      /* start scanning j */
	      j -> d = GREY;
	      buckets[j-nodes].firstActive = i;
	      i = j;
	      break;
	    }
	    else
	      if ( j -> d == GREY ) {
		/* find minimum flow on the cycle */
		delta = a -> resCap;
		while ( 1 ) {
		  delta = min ( delta, j -> current -> resCap );
		  if ( j == i )
		    break;
		  else
		    j = j -> current -> head;
		}

		/* remove delta flow units */
		j = i;
		while ( 1 ) {
		  a = j -> current;
		  a -> resCap -= delta;
		  a -> rev -> resCap += delta;
		  j = a -> head;
		  if ( j == i )
		    break;
		}
	  
		/* backup DFS to the first saturated arc */
		restart = i;
		for ( j = i -> current -> head; j != i; j = a -> head ) {
		  a = j -> current;
		  if (( j -> d == WHITE ) || ( a -> resCap == 0 )) {
		    j -> current -> head -> d = WHITE;
		    if ( j -> d != WHITE )
		      restart = j;
		  }
		}
	  
		if ( restart != i ) {
		  i = restart;
		  i->current++;
		  break;
		}
	      }
	  }
	}

	if (i->current == (i+1)->first) {
	  /* scan of i complete */
	  i -> d = BLACK;
	  if ( i != source ) {
	    if ( bos == NULL ) {
	      bos = i;
	      tos = i;
	    }
	    else {
	      i -> bNext = tos;
	      tos = i;
	    }
	  }

	  if ( i != r ) {
	    i = buckets[i-nodes].firstActive;
	    i->current++;
	  }
	  else
	    break;
	}
      } while ( 1 );
    }


  /* return excesses */
  /* note that sink is not on the stack */
  if ( bos != NULL ) {
    for ( i = tos; i != bos; i = i -> bNext ) {
      a = i -> first;
      while ( i -> excess > 0 ) {
	if (( cap[a - arcs] == 0 ) && ( a -> resCap > 0 )) {
	  if (a->resCap < i->excess)
	    delta = a->resCap;
	  else
	    delta = i->excess;
	  a -> resCap -= delta;
	  a -> rev -> resCap += delta;
	  i -> excess -= delta;
	  a -> head -> excess += delta;
	}
	a++;
      }
    }
    /* now do the bottom */
    i = bos;
    a = i -> first;
    while ( i -> excess > 0 ) {
      if (( cap[a - arcs] == 0 ) && ( a -> resCap > 0 )) {
	if (a->resCap < i->excess)
	  delta = a->resCap;
	else
	  delta = i->excess;
	a -> resCap -= delta;
	a -> rev -> resCap += delta;
	i -> excess -= delta;
	a -> head -> excess += delta;
      }
      a++;
    }
  }
}


/* gap relabeling */

int gap (bucket *emptyB)
{

  bucket *l;
  node  *i; 
  long  r;           /* index of the bucket before l  */
  int   cc;          /* cc = 1 if no nodes with positive excess before
		      the gap */

  gapCnt ++;
  r = ( emptyB - buckets ) - 1;

  /* set labels of nodes beyond the gap to "infinity" */
  for ( l = emptyB + 1; l <= buckets + dMax; l ++ ) {
    // this does nothing for high level selection 
    // but is needed for wave initialization
#ifdef WAVE_INIT
    for (i = l -> firstActive; i != sentinelNode; i = i -> bNext) {
      i -> d = n;
      gNodeCnt++;
    }
    l -> firstActive = sentinelNode;
#endif
    for ( i = l -> firstInactive; i != sentinelNode; i = i -> bNext ) {
      i -> d = n;
      gNodeCnt ++;
    }

    l -> firstInactive = sentinelNode;
  }

  cc = ( aMin > r ) ? 1 : 0;

  dMax = r;
  aMax = r;

  return ( cc );

}

/*--- relabelling node i */

long relabel (node *i)
{

  node  *j;
  long  minD;     /* minimum d of a node reachable from i */
  arc   *minA;    /* an arc which leads to the node with minimal d */
  arc   *a;

  assert(i->excess > 0);

  relabelCnt++;
  workSinceUpdate += BETA;

  i->d = minD = n;
  minA = NULL;

  /* find the minimum */
  forAllArcs(i,a) {
    workSinceUpdate++;
    if (a -> resCap > 0) {
      j = a -> head;
      if (j->d < minD) {
	minD = j->d;
	minA = a;
      }
    }
  }

  minD++;
      
  if (minD < n) {

    i->d = minD;
    i->current = minA;

    if (dMax < minD) dMax = minD;

  } /* end of minD < n */
      
  return ( minD );

} /* end of relabel */


/* discharge: push flow out of i until i becomes inactive */

void discharge (node *i)
{

  node  *j;                 /* sucsessor of i */
  long  jD;                 /* d of the next bucket */
  bucket *lj;               /* j's bucket */
  bucket *l;                /* i's bucket */
  arc   *a;                 /* current arc (i,j) */
  cType  delta;
  arc *stopA;

  assert(i->excess > 0);
  assert(i != sink);
  do {

    jD = i->d - 1;
    l = buckets + i->d;

    /* scanning arcs outgoing from  i  */
    for (a = i->current, stopA = (i+1)->first; a != stopA; a++) {
      if (a -> resCap > 0) {
	j = a -> head;

	if (j->d == jD) {
	  pushCnt ++;
	  if (a->resCap < i->excess)
	    delta = a->resCap;
	  else
	    delta = i->excess;
	  a->resCap -= delta;
	  a->rev->resCap += delta;

	  if (j != sink) {

	    lj = buckets + jD;

	    if (j->excess == 0) {
	      /* remove j from inactive list */
	      iDelete(lj,j);
	      /* add j to active list */
	      aAdd(lj,j);
	    }
	  }

	  j -> excess += delta;
	  i -> excess -= delta;
	  
	  if (i->excess == 0) break;

	} /* j belongs to the next bucket */
      } /* a  is not saturated */
    } /* end of scanning arcs from  i */

    if (a == stopA) {
      /* i must be relabeled */
      relabel (i);

      if (i->d == n) break;
      if ((l -> firstActive == sentinelNode) && 
	  (l -> firstInactive == sentinelNode)
	  )
	gap (l);

      if (i->d == n) break;
    }
    else {
      /* i no longer active */
      i->current = a;
      /* put i on inactive list */
      iAdd(l,i);
      break;
    }
  } while (1);
}


// go from higher to lower buckets, push flow
void wave() {

  node   *i;
  bucket  *l;

  for (l = buckets + aMax; l > buckets; l--) {
    for (i = l->firstActive; i != sentinelNode; i = l->firstActive) {
      aRemove(l,i);

      assert(i->excess > 0);
      discharge (i);

    }
  }
}


/* first stage  -- maximum preflow*/

void stageOne ( )

{

  node   *i;
  bucket  *l;             /* current bucket */


#if defined(INIT_UPDATE) || defined(OLD_INIT) || defined(WAVE_INIT)
  globalUpdate ();
#endif

  workSinceUpdate = 0;

#ifdef WAVE_INIT
  wave();
#endif  

  /* main loop */
  while ( aMax >= aMin ) {
    l = buckets + aMax;
    i = l->firstActive;

    if (i == sentinelNode)
      aMax--;
    else {
      aRemove(l,i);

      assert(i->excess > 0);
      discharge (i);

      if (aMax < aMin)
	break;

      /* is it time for global update? */
      if (workSinceUpdate * globUpdtFreq > nm) {
	globalUpdate ();
	workSinceUpdate = 0;
      }

    }
    
  } /* end of the main loop */
    
  flow = sink -> excess;

} 

/* 
   void scc(): Compute the strongly connected components of the residual graph  
   and builds the scc graph of this compontents. This graph is used to find all min s-t-cuts.

   Use the fieds as:
      i->d is used for dfs labels (WHITE, GREY, BLACK)
      i->bNext is used for node ordering by finishing times
      buckets[i-nodes]->firstActive is used for DSF tree

   Uses the algorithm for scc described in Cormen, et al. book.
*/

/* add an edge and its reverse to the scc graph */
void scc_add_edge(int i, int j, scc_arc*& arc)
{
  arc->head = scc_nodes + j;
  (arc+1)->head = scc_nodes + i;

  arc->next = scc_nodes[i].first_edge;
  scc_nodes[i].first_edge = arc; 
  
  (arc+1)->next = scc_nodes[j].first_rev_edge;
  scc_nodes[j].first_rev_edge = arc + 1;

  // if (j != scc_source_index)
  scc_nodes[i].outdegree++;
  
  // arc_current->rev = arc_current+1;
  // (arc_current+1)->rev = arc_current;

  // arc_current->reverse = false;
  // (arc_current+1)->reverse = true;

  arc += 2;
}

void scc()
{
  // used i->d for final finishing times
  node *i;  // current node being scanned 
  node *j;  // neighbors of i
  node *root; // root of a DFS tree
  node *prev_finished = NULL; // previous finished node by the DFS
  arc *a;
  typedef boost::unordered_set<std::pair<int,int>> edge_set_t;
  
  // initialize node info
  forAllNodes(i) {
    i->d = WHITE;
    buckets[Index(i)].firstActive = sentinelNode;
    i->current = i->first;
  }
  // start of the scc algorithm: first DFS
  // order the nodes 

  // cout << "entering DFS 1" << endl;

  forAllNodes(i)  // DFS 1
    if (i->d == WHITE)  {
      root = i;
      root->d = GREY;
      do {
	for ( ; i->current != (i+1)->first; i->current++) {
	  a = i->current;
	  if (a->resCap > 0) { // arc of the residual graph
	    j = a->head;
	    if ( j->d == WHITE ) {
	      // continue the DFS on j 
	      j->d = GREY;
	      buckets[Index(j)].firstActive = i; // add a tree edge
	      i = j;
	      break;
	    } // else if (j->d == GREY) "a is a back edge" 
	  }	    
	} // end for ( ; i->current != ...
	// finish the search from i
	if (i->current == (i+1)->first) {
	  i->d = BLACK;	
	  i->bNext = prev_finished;
	  prev_finished = i;
	
	  if ( i != root) { // go up the DFS tree
	    i = buckets[Index(i)].firstActive;
	    i->current++;
	  } else // end of a tree traversal, go to the next
	    break;
	}  
      } while (1); 
    } // if (i->d == WHITE) and end of forAllNodes
  
  // the second DFS is done on the reverse residual graph

  // Build the scc graph.
  // The root of each dfs tree represents one vertex.
  // Edges of the scc graph will be represented by ....

  // initialize node info
  forAllNodes(i) {
    i->d = WHITE;
    buckets[Index(i)].firstActive = sentinelNode;
    i->current = i->first;
  }
  int scc_num = -1;  // component number 
  // use vertex ordering from the first dfs
  // use i->bPrev to point to the next node in the scc
  // user i->excess to store the scc number

  // set to store edges without repetition
  edge_set_t edge_set;
  vector<node*> tree_roots;
  // std::unordered_set<int> edge_set;

  // cout << "entering DFS 2" << endl;  
  // cout << "source/sink " << Index(source) << " " << Index(sink) << endl;
  for (i = prev_finished; i != NULL; i = i->bNext)
    if (i->d == WHITE)  {
      root = i;
      root->d = GREY;
      root->excess = ++scc_num;
      node* prev_in_component = root;
      scc_nodes[scc_num].size = 1;
      tree_roots.push_back(root);
      root->bPrev = 0;
      // cout << "root: " << Index(root);
      if (root == source)
	scc_source_index = scc_num;
      else if (root == sink)
	scc_sink_index = scc_num;
      do {
	for ( ; i->current != (i+1)->first; i->current++) {
	  a = i->current;
	  // cout << "seeing edge " << Index(i) << " - " << Index(a->head)  
	  //     << "  rev->resCap = " << a->rev->resCap << endl;
	  if (a->rev->resCap > 0) { // traversing the arc reversed residual graph
	    j = a->head;
	    if ( j->d == WHITE ) {
	      // continue the DFS on j 
	      j->d = GREY;
	      buckets[Index(j)].firstActive = i; // add a tree edge
	      j->excess = scc_num;
	      prev_in_component->bPrev = j; // link the vertices in the ssc
	      prev_in_component = j;
	      j->bPrev = 0;
	      scc_nodes[scc_num].size++;
	      i = j;
	      // cout << "visited " << Index(j);
	      if (i == source)
		scc_source_index = scc_num;
	      else if (i == sink)
		scc_sink_index = scc_num;
	      break;
	    } else  // if (j->d == GREY) "a is a back edge" 
	      if (j->d == BLACK && j->excess != i->excess) {  // found a (reversed) edge to other scc
		// note that i->excess and j->excess had been already defined
		// cout << "adding to hash " << Index(j) << "-" << Index(i) << endl;
		edge_set.insert(pair<int,int> (j->excess, i->excess)); // insert the reverse edge (j,i)
	      }
	  }
	} // end for ( ; i->current != ...
	if (i->current == (i+1)->first) {
	// finish the search from i
	  i->d = BLACK;	
	  if ( i != root) { // go up the DFS tree
	    i = buckets[Index(i)].firstActive;
	    i->current++;
	  } else // end of a tree traversal, go to the next
	    break;
	}  
      } while(1);
    } // if (i->d == WHITE) and end of forAllNodes 

  // build the scc graph (it is a d.a.g.)
  scc_n = scc_num+1;
  scc_m = edge_set.size();
  scc_arc *arc_current = scc_arcs;
  scc_sentinelNode = scc_nodes + scc_n;
   
  // unordered_set<std::pair<int,int>>::iterator it;  
  scc_node* u; 
  int ii;
  for (ii = 0, u = scc_nodes; u != scc_sentinelNode; u++, ii++)
    {
      u->first_edge = u->first_rev_edge = 0;
      u->first_vertex = tree_roots[ii];
      u->outdegree = 0;
    }
  // cout << "Building the scc graph" << endl;

  //  for (it = edge_set.begin(); it != edge_set.end(); it++)
  // for (const std::pair<int,int>& edge: edge_set)
  BOOST_FOREACH(edge_set_t::value_type edge, edge_set) 
    {
      //   add edge ( *it.first, *it.secod)  to the scc graph
      int i = edge.first;
      int j = edge.second;
      scc_add_edge(i, j, arc_current);
      // cout << "adding edge: " << i << " " << j << endl;
    } 
  scc_source = &scc_nodes[scc_source_index];
  scc_sink = &scc_nodes[scc_sink_index];
  // DONE! Now go to enumerate all oriented cuts of the scc dag!
  // cout << "Done building the scc graph" << endl;
}

void print_scc_graph()
{
  int i;
  scc_arc* a;
  
  cout << "scc graph: "<< endl;
  cout << "n = " << scc_n << " m=" << scc_m  << endl;
  cout << "source = " << scc_source - scc_nodes << endl;
  cout << "sink = " << scc_sink - scc_nodes << endl;
  for (i=0; i < scc_n; i++)
    {
      cout << i << ": ";
      for (node* u = scc_nodes[i].first_vertex; u != 0; u = u->bPrev)
	cout << nNode(u) << " ";
      cout << endl;
    }
      
  for (i=0; i < scc_n; i++)
    for (a = scc_nodes[i].first_edge; a != 0; a = a->next)
	cout << i << " " << a->head-scc_nodes << endl;
}

// - decrease outdegrees of the inverse neighbors of pivot
// - called when pivot is put in S
void decrease_outdegrees(scc_node* pivot, pivot_set_t& pivot_set)
{
  scc_node* v;
  scc_arc* a; 
  for (a = pivot->first_rev_edge; a != 0; a = a->next)
    {
      v = a->head;
      // cout << "pivot " << pivot-scc_nodes << "  decreasing " << v-scc_nodes << endl;
      assert(v->outdegree > 0);
      if(--v->outdegree == 0 && !v->in_S && !v->in_T)  // insert in pivot set
	pivot_set.insert(v);
    }  
}

// - increase outdegrees of the inverse neighbors of pivot
// - called when pivot is removed from S (and put in T)
void increase_outdegrees(scc_node* pivot, pivot_set_t& pivot_set)
{
  scc_node* v;
  scc_arc* a; 
  for (a = pivot->first_rev_edge; a != 0; a = a->next)
    {
      v = a->head;
      if(++v->outdegree == 1) // remove v from the pivot_set
	pivot_set.erase(v);
    }
}

// auxiliary dfs for enumerate_all_st_cuts
// visit reachable node from the pivot, set in_S true and 
// add nodes to the stack in_S
bool reachable_from_node(scc_node* source, vector<scc_node*>& in_S, pivot_set_t& pivot_set) 
{
  scc_node* u;
  scc_node* v;
  scc_arc* a; 

  source->in_S = true;
  in_S.push_back(source);
  S_size += source->size;
  decrease_outdegrees(source, pivot_set);
  u = source;
  do
    {
      for ( ; u->current != 0; u->current = u->current->next)
	{
	  a = u->current; // visiting arc a
	  v = a->head;    // reached node v
	  if (!v->in_S)
	    {
	      v->in_S = true;
	      in_S.push_back(v);
	      S_size += v->size;
	      decrease_outdegrees(v, pivot_set);
	      v->current = v->first_edge;
	      v->pred = u;
	      u = v;
	      break;
	    }
	}
      if (u->current == 0) {
	if (u != source) {
	  u = u->pred;
	  u->current = u->current->next; // next edge
	} else
	  break;
      }
    } while (1);
  return true;
}

/* use the reverse graph to find all nodes that reach the pivot */
bool reaches_node(scc_node* destination)
{
  scc_node* u;
  scc_node* v;
  scc_arc* a; 

  /* for (u = scc_nodes; u != scc_sentinelNode; u++) {
    u->rev_current = u->first_rev_edge; 
    } */
  
  // pivot->visited = true;
  destination->in_T = true;
  destination->rev_current = destination->first_rev_edge;
  u = destination;
  do
    {
      for ( ; u->rev_current != 0; u->rev_current = u->rev_current->next)
	{
	  a = u->rev_current; // visiting arc a
	  v = a->head;    // reached node v
	  assert(!v->in_S);
	  if (!v->in_T)
	    {
	      v->in_T = true;
	      v->rev_current = v->first_rev_edge;
	      v->pred = u;
	      u = v;
	      break;
	    }
	}
      if (u->rev_current == 0) {
	if (u != destination) {
	  u = u->pred;
	  u->rev_current = u->rev_current->next; // next edge
	} else
	  break;
      }
    } while (1);
  return true;
}

void print_cut(vector<scc_node*>& in_S, int& number_of_cuts)
{
  cout << "Cut "<< number_of_cuts << ": " ;
  for (vector<scc_node*>::iterator it = in_S.begin(); it != in_S.end(); it++) {
    cout << *it - scc_nodes << "(";
    node* v = (*it)->first_vertex;
    while (v != 0) {  // list the original vertices
      cout << Index(v)+1 << ",";
      v = v->bPrev;
    }
    cout << "), ";
  }
  cout << endl;
}

// Enumerate all directed cuts of the scc graph of the residual graph 
// because these cuts correspond to the minimum s-t-cuts of the input graph
// The enumeration algorithm:
//    if cannot find a valid pivot
//    then treats the cut (S, V-S)
//    else ENUM (S \cup {u | pivot reaches u}, T)
//         ENUM (S, T \cup { pivot }


void enumerate_all_st_cuts(vector<scc_node*>& in_S, pivot_set_t& pivot_set, vector<bool>& cut)
{
  // pick a pivot (the node that goes to S and then to T)
  scc_node* pivot = 0;

  if (number_of_cuts > 1000) // too many cuts to search -> return
    return ; 

  while (!pivot_set.empty()) {
    pivot_set_iterator u_it = pivot_set.begin();
    scc_node* u = *u_it;
    pivot_set.erase(u_it);
    if (!u->in_S && !u->in_T) { // is this test necessary?
      pivot = u;
      break;
    }
  }
      
  if (pivot) {
#ifdef VERBOSE
    cout << "Pivot: " << pivot - scc_nodes << endl;
#endif
    // add pivot to in_S 
    pivot->in_S = true;
    in_S.push_back(pivot);
    S_size += pivot->size;
    // update outdegree of reverse neighbors of pivot
    decrease_outdegrees(pivot, pivot_set);
    
    enumerate_all_st_cuts(in_S, pivot_set, cut);
    
    in_S.pop_back();  // remove pivot from S
    pivot->in_S = false;
    S_size -= pivot->size;
    // reaches_pivot(pivot); // nodes that reach the pivot go to T   
    pivot->in_T = true; // add pivot to T

    // update indegrees and the pivot_list
    increase_outdegrees(pivot, pivot_set);

    enumerate_all_st_cuts(in_S, pivot_set, cut);

    // reinsert pivot into the pivot set
    pivot_set.insert(pivot);
    pivot->in_T = false;
  } else {   // treat the cut (S, V-S)

    float balance_factor = fabs(n/2.0 - S_size);

    // cout << "Size = " << S_size << " " << balance_factor << endl;
    ++number_of_cuts;
    if (balance_factor < most_balanced_value)
      {
	most_balanced_value = balance_factor;
	for (int i = 0; i < n; i++)
	  cut[i] = true;

	for (vector<scc_node*>::iterator it = in_S.begin(); it != in_S.end(); it++) {
	  node* v = (*it)->first_vertex;
	  while (v != 0) {  // list the original vertices
      	    cut[Index(v)] = false;
	    v = v->bPrev;
	  }
	}
      }
#ifdef VERBOSE
    print_cut(in_S, number_of_cuts);
#endif
  }
} 

void find_balanced_cut(vector<bool>& cut, float& return_balance_factor)
{
  vector<scc_node*> in_S;
  pivot_set_t pivot_set;

  scc(); // build the condensed graph

#ifdef VERBOSE
  print_scc_graph();
#endif

  number_of_cuts = 0;
  S_size = 0;
  most_balanced_value = n+1;
  
  // initialize S and T
  scc_node* u;
  for (u = scc_nodes; u != scc_sentinelNode; u++) {
      u->in_S = false;
      u->in_T = false;
  }

  // include source and all nodes reachable from source in S
  reachable_from_node(scc_source, in_S, pivot_set);

  // include sink and all nodes that reaches the sink in T
  reaches_node(scc_sink);

  // nodes with outdegree 0 in G[\overline{S}] go to the candidate set of pivots
  /* for (u = scc_nodes; u != scc_sentinelNode; u++)
    if (u->outdegree == 0 && !u->in_S && !u->in_T) // insert u into the pivot set
    pivot_set.insert(u); */

  // recursively enumerate all cuts that separates S from T
  enumerate_all_st_cuts(in_S, pivot_set, cut);
  
  float mm = n/2.0;
  return_balance_factor = (mm-1-most_balanced_value)/(mm-1);
}

double hi_pr (Edgelist& G, vector<bool>& cut, int balanced, double& return_time, int& return_num_of_cuts, float& return_balance_factor)
{
#if (defined(PRINT_FLOW) || defined(CHECK_SOLUTION))
  node *i;
  arc *a;
#endif

#ifdef PRINT_FLOW
  long ni, na;
#endif
  node *j;
  int  cc;
  /* #ifdef CHECK_SOLUTION */
  excessType sum;
  bucket *l;
  /* #endif */

  timer1();

  globUpdtFreq = GLOB_UPDT_FREQ;
  /* globUpdtFreq = (float) atof(argv[1]); */

  parse_edgelist(G, &n, &m, &nodes, &arcs, &cap, &source, &sink, &nMin );

  cc = allocDS();
  if ( cc ) { fprintf ( stderr, "Allocation error\n"); exit ( 1 ); }

  init();
  stageOne ( );

  // t2 = timer() - t2;

#ifndef CUT_ONLY
  stageTwo ( );

  /* printf ("c time:        %10.2f\n", t); */

#endif

  /* printf ("c cut tm:      %10.2f\n", t2); */

#ifdef CHECK_SOLUTION

  /* check if you have a flow (pseudoflow) */
  /* check arc flows */
  forAllNodes(i) {
    forAllArcs(i,a) {
      if (cap[a - arcs] > 0) /* original arc */
	if ((a->resCap + a->rev->resCap != cap[a - arcs]) 
	    || (a->resCap < 0)
	    || (a->rev->resCap < 0)) {
	  printf("ERROR: bad arc flow\n");
	  exit(2);
	}
    }
  }

  /* check conservation */
  forAllNodes(i)
    if ((i != source) && (i != sink)) {
#ifdef CUT_ONLY
      if (i->excess < 0) {
	printf("ERROR: nonzero node excess\n");
	exit(2);
      }
#else
      if (i->excess != 0) {
	printf("ERROR: nonzero node excess\n");
	exit(2);
      }
#endif

      sum = 0;
      forAllArcs(i,a) {
	if (cap[a - arcs] > 0) /* original arc */
	  sum -= cap[a - arcs] - a->resCap;
	else
	  sum += a->resCap;
      }

      if (i->excess != sum) {
	printf("ERROR: conservation constraint violated\n");
	exit(2);
      }
    }

  /* check if mincut is saturated */
  aMax = dMax = 0;
  for (l = buckets; l < buckets + n; l++) {
    l->firstActive = sentinelNode;
    l->firstInactive = sentinelNode;
  }
  globalUpdate();
  if (source->d < n) {
    printf("ERROR: the solution is not optimal\n");
    exit(2);
  }

  printf("c\nc Solution checks (feasible and optimal)\nc\n");
#endif

#ifdef PRINT_STAT
    printf ("c pushes:      %10ld\n", pushCnt);
    printf ("c relabels:    %10ld\n", relabelCnt);
    printf ("c updates:     %10ld\n", updateCnt);
    printf ("c gaps:        %10ld\n", gapCnt);
    printf ("c gap nodes:   %10ld\n", gNodeCnt);
    printf ("c\n");
#endif

#ifdef PRINT_FLOW
    printf ("c flow values\n");
    forAllNodes(i) {
      ni = nNode(i);
      forAllArcs(i,a) {
	na = nArc(a);
	if ( cap[na] > 0 )
	  printf ( "f %7ld %7ld %12ld\n",
		  ni, nNode( a -> head ), cap[na] - ( a -> resCap )
		  );
      }
    }
    printf("c\n");
#endif
    
  aMax = dMax = 0;
  for (l = buckets; l < buckets + n; l++) {
    l->firstActive = sentinelNode;
    l->firstInactive = sentinelNode;
  }
  globalUpdate(); /* moved from inside #ifdef PRINT_CUT */

#ifdef PRINT_CUT
  printf ("c nodes on the sink side\n");
  forAllNodes(j)
    if (j->d < n)
      printf("c %ld\n", nNode(j));
#endif

  if (!balanced) {
    int cut_size = 0;
    for (int i = 0; i < n; i++)
      cut[i] = false;
    forAllNodes(j)
      if (j->d < n) {
	cut[Index(j)] = true;
	cut_size++;
      }
    return_num_of_cuts = 1;
 
    float mm = n/2.0;
    return_balance_factor = (mm-1-fabs(mm-cut_size))/(mm-1);
  } else {
    find_balanced_cut(cut, return_balance_factor);
    return_num_of_cuts = number_of_cuts;
  }

#ifdef VERBOSE1
    cout << "hi_pr cut: " ;
    for (int i = 0; i < n; i++)
      if (cut[i])
	cout << i+1 << " ";
    cout << endl;
#endif

  timer2();
  return_time  = getRealTime();
  // sendbuf[1] = getUserTime();
  // sendbuf[2] = getSystemTime();
  
  // return stats
  // cut << fixed;
  // cout << setprecision(8) << time_per_cut << " " << return_num_of_cuts << endl;

  return flow;
}
